/* 
 * $Id: bgp_dump.c,v 1.21 1999/01/28 03:27:49 labovit Exp $
 */

#include <mrt.h>
#include <bgp.h>
#include <sys/wait.h>

/* local functions */
static void dump_binary_bgp_view (int fd, view_t *view); 

void
bgp_write_mrt_msg (bgp_peer_t *peer, enum MRT_MSG_BGP_TYPES bgptype, 
		   u_char *buf, int length) {
    u_char tmpx[MAX_MSG_SIZE], *cp = tmpx;
    int len = length;
    char sync[MAXLINE], *xp = sync;
    int synclen = 0;
    time_t now;
    enum MRT_MSG_TYPES msgtype = MSG_PROTOCOL_BGP;
    int viewno = peer->view;
#if defined(_REENTRANT) && defined(HAVE_LOCALTIME_R)
    struct tm my_tm;
#endif
    pid_t pid;

    if (BGP->dump_update_form == NULL && BGP->dump_route_form[viewno] == NULL)
	return;

    pthread_mutex_lock (&BGP->mutex_lock);
    time (&now);

    if (BGP->dump_route_form[viewno]) {
      /* don't dump on startup -- the tables won't be very interesting */
      if (BGP->dump_route_time[viewno] == 0)
	BGP->dump_route_time[viewno] = now;

        if (BGP->dump_route_time[viewno] + BGP->dump_route_interval[viewno] 
	    <= now) {

	  char name[MAXLINE];
	  struct tm *tm;
	  int fd;
	  uii_connection_t uii;
	  static int nerror = 0;

	  tm = localtime (&now);
	  strftime (name, sizeof (name), BGP->dump_route_form[viewno], tm);
	  if ((fd = open (name, (O_CREAT | O_APPEND | O_WRONLY), 0666)) < 0) {
	    if (nerror++ <= 0) {
	      trace (TR_ERROR, MRT->trace, 
		     "can not open dump file: %s (%s)\n",
		     name, strerror (errno));
	    }
	    else if (nerror >= 100) {
	      trace (TR_ERROR, MRT->trace, 
		     "give up opening dump file: %s (%s)\n",
		     name, strerror (errno));
	      Delete (BGP->dump_route_form[viewno]);
	      BGP->dump_route_form[viewno] = NULL;
	    }
	    BGP->dump_route_time[viewno] = now;
	  }
	  else {
	    trace (NORM, MRT->trace, "Dumping routing table to disk\n");

#ifndef HAVE_PTHREAD_H
	    /* it was taking too long to do the dumps, hopefully forking
	     * will be faster...
	     */

	    BGP->dump_route_time[viewno] = now;

	    if ((pid = fork ()) == -1) {
	      perror ("fork");
	      exit (1);
	    }
	    else if (pid > 0) {	/* parent -- main process */
	      goto after_fork;
	    }

	    /*
	     * fork again so that not being a process group leader 
	     */
	    if ((pid = fork ()) == -1) {
	      perror ("fork");
	      exit (1);
	    }
	    else if (pid > 0) {	/* parent of second fork */
	      exit (0);
	    }
#endif /* HAVE_PTHREAD_H */

	    lseek (fd, 0, SEEK_END);
	    memset (&uii, 0, sizeof (uii));
	    uii.sockfd = -1;
	    uii.sockfd_out = fd;
	    if (bgptype == MSG_BGP_UPDATE) {
	      if (BGP->dump_route_type == DUMP_ASCII) {
		bgp_dump_view (&uii, BGP->views[viewno], NULL, NULL, NULL);
		uii_send_bulk_data (&uii);
	      }
	      else
		dump_binary_bgp_view (fd, BGP->views[viewno]);
	    }
	    else {
	      view_open (BGP->views[viewno]);
	      if (BGP->dump_route_type == DUMP_ASCII) {
		bgp_dump_view (&uii, BGP->views[viewno], 
			       NULL, NULL, NULL);
		uii_send_bulk_data (&uii);
	      }
	      else
		dump_binary_bgp_view (fd, BGP->views[viewno]);
	      view_close (BGP->views[viewno]);
	    }
	    close (fd);
	    BGP->dump_route_time[viewno] = now;

	    /* set sync information */
	    if (BGP->dump_update_form) {
	      BGP_PUT_SHORT (viewno, xp);
	      strcpy (xp, name);
	      synclen = sizeof (short) + strlen (name) + 1 /* for zero */;
	    }
	  }
#ifndef HAVE_PTHREAD_H
	  /* we are the child, exit here */
	  exit (0);
#endif /* HAVE_PTHREAD_H */
	}
    }
    
    
after_fork: /* parent has skipped dump */

#ifndef HAVE_PTHREAD_H 
    if (waitpid (pid, NULL, 0) != pid)
      perror ("waitpid error");
#endif /* HAVE_PTHREAD_H */

    if (BGP->dump_update_form) {
        static io_t *IO = NULL;

        if (IO == NULL) {
	    IO = New_IO (NULL);
        }
  
        if (BGP->dump_update_time == 0 ||
	   (BGP->dump_update_interval && 
		BGP->dump_update_time + BGP->dump_update_interval <= now)) {

	    char name[MAXLINE];
	    struct tm *tm;
	    static int nerror = 0;

#if defined(_REENTRANT) && defined(HAVE_LOCALTIME_R)
	    tm = localtime_r (&now, &my_tm);
#else
	    tm = localtime (&now);
#endif
            strftime (name, sizeof (name), BGP->dump_update_form, tm);
	    if (io_set (IO, IO_OUTFILE, name, NULL) < 0) {
		if (nerror++ <= 0) {
		    trace (TR_ERROR, MRT->trace, 
		           "can not open dump file: %s (%s)\n",
		           name, strerror (errno));
		}
		else if (nerror >= 100) {
		    trace (TR_ERROR, MRT->trace, 
		           "give up opening dump file: %s (%s)\n",
		           name, strerror (errno));
		    Delete (BGP->dump_update_form);
    		    BGP->dump_update_form = NULL;
		}
		BGP->dump_update_time = now;
    		pthread_mutex_unlock (&BGP->mutex_lock);
		return;
	    }
	    else {
		nerror = 0;
		BGP->dump_update_time = now;
	    }
	}

        /* from */
        BGP_PUT_SHORT (peer->gateway->AS, cp);

#ifdef HAVE_IPV6
        if (peer->gateway->prefix->family == AF_INET6) {
            memcpy (cp, prefix_tochar (peer->gateway->prefix), 16);
            cp += 16;
        }
        else
#endif /* HAVE_IPV6 */
        BGP_PUT_NETLONG (prefix_tolong (peer->gateway->prefix), cp);

	/*
	 * To be compatible with Craig's code, AS and addr for 'to' 
	 * should not be included in this MRT message at the moment. 
	 *   -- masaki
	 */
	if (bgptype != MSG_BGP_STATE_CHANGE) {

	  /* to */
	  BGP_PUT_SHORT (BGP->my_as, cp);
#ifdef HAVE_IPV6
	  if (peer->gateway->prefix->family == AF_INET6) {
            if (peer->local_addr) {
	      assert (peer->local_addr->family == AF_INET6);
	      memcpy (cp, prefix_tochar (peer->local_addr), 16);
	    }
	    else {
	      memset (cp, 0, 16);
	    }
            cp += 16;
	  }
	  else
#endif /* HAVE_IPV6 */
	    if (peer->local_addr)
	      BGP_PUT_NETLONG (prefix_tolong (peer->local_addr), cp);
	    else
	      BGP_PUT_NETLONG (0L, cp);
	}

        memcpy (cp, buf, length);
        len += (cp - tmpx);

#ifdef HAVE_IPV6
        if (peer->gateway->prefix->family == AF_INET6) {
	  if (BIT_TEST (peer->options, BGP_BGP4PLUS_01))
	    msgtype = MSG_PROTOCOL_BGP4PLUS_01;
	  else
	    msgtype = MSG_PROTOCOL_BGP4PLUS;
        }
#endif /* HAVE_IPV6 */


#ifdef HAVE_PTHREAD_H
 	/* flush sync info */
	if (synclen > 0)
	  io_write (IO, 0, msgtype, MSG_BGP_SYNC, synclen, (char *)sync);
#endif /* HAVE_PTHREAD_H */

        io_write (IO, 0, msgtype, bgptype, len, (char *)tmpx);
    }
    pthread_mutex_unlock (&BGP->mutex_lock);
}


void
bgp_write_status_change (bgp_peer_t *peer, u_short state) {
    u_char buf[2+2], *cp = buf;

    BGP_PUT_SHORT (peer->state, cp); /* previous */
    BGP_PUT_SHORT (state, cp); /* new */
    bgp_write_mrt_msg (peer, MSG_BGP_STATE_CHANGE, buf, 4);
}



/* dump_binary_bgp_view 
 * 4 time stamp | 2 type | 2 subtype (familly) | 4 length
 * 2  view # | 
 *
 *  4 prefix | 1 mask | status (VRTS_SUPPRESS) | 4 time originated |
 *  4 len | attributes 
 * 
 */
static void dump_binary_bgp_view (int fd, view_t *view) {
  bgp_route_head_t *rt_head;
  bgp_route_t *route;
  u_char buffer[MAX_MSG_SIZE];
  time_t now;
  u_char *cp, *end, *start_attr;
  u_char *p_save_len, *p_total_attrib_len;
  int len;
  short seq_num = 0;
  
  time (&now);

  cp = buffer;
  end = buffer + MAX_MSG_SIZE;

  BGP_PUT_LONG (now, cp);		/* time */
  BGP_PUT_SHORT (MSG_TABLE_DUMP, cp);
  BGP_PUT_SHORT (0, cp);		/* subtype. maybe family ? */
  p_save_len = cp;
  BGP_PUT_LONG (0, cp);			/* skip length for now */

  BGP_PUT_SHORT (23, cp);		/* view */
  BGP_PUT_SHORT (seq_num, cp);		/* sequence number */
  seq_num++;

  VIEW_RADIX_WALK (view, rt_head) {

    if ((end - cp) < 500) {
      len = cp - buffer - 12;
      BGP_PUT_LONG (len, p_save_len);
      cp = buffer;
      write (fd, cp, len+12);
      trace (NORM, MRT->trace, 
	     "Dumping routing table (%d bytes) to disk\n", len+12);
      
      cp = buffer;
      memset (cp, 0, MAX_MSG_SIZE);
      
      BGP_PUT_LONG (now, cp);
      BGP_PUT_SHORT (MSG_TABLE_DUMP, cp);
      BGP_PUT_SHORT (0, cp);	/* subtype. maybe family ? */
      p_save_len = cp;
      BGP_PUT_LONG (0, cp);	/* skip length for now */

      BGP_PUT_SHORT (0, cp);	/* view */
      BGP_PUT_SHORT (seq_num, cp); /* view */
      seq_num++;
    }
    
    
    /* 
     * BGP Routes 
     */
    LL_Iterate (rt_head->ll_routes, route) {
      assert (rt_head->active);

      BGP_PUT_ADDR (prefix_tochar (rt_head->prefix), cp);
      BGP_PUT_BYTE (rt_head->prefix->bitlen, cp);

/* XXX I need to check this Craig's code.
   Probably it's OK putting state_bits but it's u_long. masaki */
      /* status */
      if (BIT_TEST (rt_head->state_bits, VRTS_SUPPRESS))
	BGP_PUT_BYTE (VRTS_SUPPRESS, cp);
      else if (BIT_TEST (rt_head->state_bits, VRTS_DELETE))
	BGP_PUT_BYTE (VRTS_DELETE, cp);
      else
	BGP_PUT_BYTE (1, cp); /* active ? */

      /* age, or really time last changed */
      BGP_PUT_LONG (route->time, cp);

      /* peer */
      BGP_PUT_ADDR (prefix_tochar (route->attr->gateway->prefix), cp);
      BGP_PUT_SHORT (route->attr->gateway->AS, cp);

      /* skip attribute length for now */
      p_total_attrib_len = cp;
      BGP_PUT_SHORT (0, cp);
      start_attr = cp;

      cp = bgp_add_attributes (cp, end - cp, route->attr, MRT->trace);


      BGP_PUT_SHORT (cp - start_attr, p_total_attrib_len);
    }
  }
  VIEW_RADIX_WALK_END;

  /* does this include header ? */
  len = cp - buffer - 12;
  BGP_PUT_LONG (len, p_save_len);
  cp = buffer;
  write (fd, cp, len + 12);
  trace (NORM, MRT->trace, 
	 "Dumping routing table (%d bytes) to disk\n", len+12);
}





